/*
 Tux Droid - Dynamic time warping
 Copyright (C) 2007 C2ME Sa <remi.jocaille@c2me.be>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define  MIN(a, b)       (((a) < (b)) ? (a) : (b))
#define  MAX(a, b)       (((a) > (b)) ? (a) : (b))
#define  DIST(a, b)       (((a) > (b)) ? (a - b) : (b - a))

struct m_patern
{
    float* matrix;
    unsigned int matrix_len;
};

struct m_patern p1;
struct m_patern p2;

struct m_patern read_file(char* path)
{
	FILE* fp;
	unsigned int size;
	unsigned int float_count;
	struct m_patern m;
	
	if ((fp = fopen(path, "rb")) == NULL){
      fprintf(stderr, "Cannot open file %s!\n", path);
      exit(2);
    }
	
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	float_count = (unsigned int)(size / sizeof(float));
	m.matrix = (float *)malloc(size);
	fseek(fp, 0, SEEK_SET);
	fread(m.matrix, sizeof(float), float_count, fp);
	fclose(fp);
	m.matrix_len = float_count;
	return m;
}

float matrix_dist(struct m_patern m1, struct m_patern m2)
{
	float dist;
	float coeff;
	int i;
	int j;
	
	dist = 0;
	
	if (m1.matrix_len > m2.matrix_len)
	{
		coeff = m1.matrix_len / m2.matrix_len;
		for (i=0; i<m1.matrix_len; i++)
		{
			j = (int)(coeff * i);
			if (j >= m2.matrix_len)
				j = m2.matrix_len - 1;
			dist += DIST(m1.matrix[i], m2.matrix[j]);
		}
		dist /= m1.matrix_len;
	}else{
		coeff = m2.matrix_len / m1.matrix_len;
		for (i=0; i<m2.matrix_len; i++)
		{
			j = (int)(coeff * i);
			if (j >= m1.matrix_len)
				j = m1.matrix_len - 1;
			dist += DIST(m2.matrix[i], m1.matrix[j]);
		}
		dist /= m2.matrix_len;
	}
	return dist;
}

float DTW(struct m_patern m1, struct m_patern m2)
{
	int i, k, j, l;
	unsigned int long_ref;
	unsigned int long_test;
	int a, b;
	float dist_local[m1.matrix_len];
	float dist_acc0[m1.matrix_len];
	float dist_acc1[m1.matrix_len];
	
	for (i=0; i<m1.matrix_len; i++){
		dist_local[i] = 0.0;
		dist_acc0[i] = 0.0;
		dist_acc1[i] = 0.0;
	}
	
	long_ref = m1.matrix_len;
	long_test = m2.matrix_len;
	
	if (long_ref == 0)
		return 9999999999.0;
	if (long_test == 0)
		return 9999999999.0;
	
	dist_local[0] = DIST(m1.matrix[0], m2.matrix[0]);
	dist_acc1[0] = dist_local[0];
	
	for (i=1; i<long_ref; i++)
		dist_acc1[i] = 999999999.0;
	
	for (k=1; k<long_test; k++)
	{
		for (j=0; j<long_ref; j++)
			dist_acc0[j] = dist_acc1[j];
		
		a = (int)((2*k) + (long_ref) - (2*long_test));
		b = (int)(-1 + (2*k));
		
		for (l=MAX(a, 0); l<MIN(b, long_ref); l++)
			dist_local[l] = DIST(m2.matrix[k], m1.matrix[l]);
		
		if (a<1)
			dist_acc1[0] = dist_local[0] + dist_acc0[0];
		else
			dist_acc1[0] = 999999999.0;
		
		for (l=1; l<(a-1); l++)
			dist_acc1[l] = 999999999.0;
		
		for (l=MAX(1, a); l<MIN(long_ref, b); l++)
			dist_acc1[l] = dist_local[l] + MIN(MIN(dist_acc1[l-1], dist_acc0[l]), dist_acc0[l-1]);
		
		for (l=(b+1); l<long_ref; l++)
			dist_acc1[l] = 999999999.0;
	}
	return dist_acc1[long_ref-1] / (long_ref + long_test);
}

int main (int argc, char* argv[])
{
	float dist;
	
	p1 = read_file(argv[1]);
	p2 = read_file(argv[2]);
	dist = DTW(p1, p2);
	printf("%f\n", dist);
	return 0;
}
